#pragma once

#include <map>
#include <string>

#include <SimoxUtility/algorithm/get_map_keys_values.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "../../core/Time.h"
#include "../../core/MemoryID.h"

#include "EntitySnapshotBase.h"
#include "detail/MaxHistorySize.h"
#include "detail/MemoryContainerBase.h"


namespace armarx::armem::base
{

    /**
     * @brief An entity over a period of time.
     *
     * An entity should be a physical thing or abstract concept existing
     * (and potentially evolving) over some time.
     *
     * Examples are:
     * - objects (the green box)
     * - agents (robot, human)
     * - locations (frige, sink)
     * - grasp affordances (general, or for a specific object)
     * - images
     * - point clouds
     * - other sensory values
     *
     * At each point in time (`EntitySnapshot`), the entity can have a
     * (potentially variable) number of instances (`EntityInstance`),
     * each containing a single `AronData` object of a specific `AronType`.
     */
    template <class _EntitySnapshotT, class _Derived>
    class EntityBase :
        public detail::MemoryContainerBase<std::map<Time, _EntitySnapshotT>, _Derived>,
        public detail::MaxHistorySize
    {
        using Base = detail::MemoryContainerBase<std::map<Time, _EntitySnapshotT>, _Derived>;

    public:

        using typename Base::DerivedT;
        using typename Base::ContainerT;

        using EntitySnapshotT = _EntitySnapshotT;
        using EntityInstanceT = typename EntitySnapshotT::EntityInstanceT;


    public:

        EntityBase()
        {
        }
        EntityBase(const std::string& name, const MemoryID& parentID = {}) :
            EntityBase(parentID.withEntityName(name))
        {
        }
        EntityBase(const MemoryID& id) :
            Base(id)
        {
        }


        EntityBase(const EntityBase& other) = default;
        EntityBase(EntityBase&& other) = default;
        EntityBase& operator=(const EntityBase& other) = default;
        EntityBase& operator=(EntityBase&& other) = default;


        virtual bool equalsDeep(const EntityBase& other) const
        {
            //std::cout << "Entity::equalsDeep" << std::endl;
            if (this->size() != other.size())
            {
                return false;
            }
            for (const auto& [key, snapshot] : this->_container)
            {
                if (not other.hasSnapshot(key))
                {
                    return false;
                }
                if (not snapshot.equalsDeep(other.getSnapshot(key)))
                {
                    return false;
                }
            }
            return true;
        }


        inline const std::string& name() const
        {
            return this->id().entityName;
        }
        inline std::string& name()
        {
            return const_cast<std::string&>(const_cast<const EntityBase*>(this)->name());
        }


        inline const std::map<Time, EntitySnapshotT>& history() const
        {
            return this->_container;
        }
        inline std::map<Time, EntitySnapshotT>& history()
        {
            return const_cast<std::map<Time, EntitySnapshotT>&>(const_cast<const EntityBase*>(this)->history());
        }


        /**
         * @brief Indicates whether a history entry for the given time exists.
         */
        bool hasSnapshot(Time time) const
        {
            return this->_container.count(time) > 0;
        }

        /**
         * @brief Get the latest timestamp.
         * @throw `armem::error::EntityHistoryEmpty` If the history is empty.
         */
        Time getLatestTimestamp() const
        {
            return getLatestItem().first;
        }

        /**
         * @brief Get all timestamps in the history.
         */
        std::vector<Time> getTimestamps() const
        {
            return simox::alg::get_keys(this->_container);
        }


        /**
         * @brief Get a snapshot.
         * @param time The time.
         * @return The snapshot, if it exists.
         *
         * @throws `armem::error::MissingEntry` If there is no such entry.
         * @throws `armem::error::MissingData` If the entry has no data.
         */
        EntitySnapshotT& getSnapshot(Time time)
        {
            return const_cast<EntitySnapshotT&>(const_cast<const EntityBase*>(this)->getSnapshot(time));
        }

        const EntitySnapshotT& getSnapshot(Time time) const
        {
            auto it = this->_container.find(time);
            if (it != this->_container.end())
            {
                return it->second;
            }
            else
            {
                throw error::MissingEntry("entity snapshot", toDateTimeMilliSeconds(time), getLevelName(), this->id().str());
            }
        }

        EntitySnapshotT& getSnapshot(const MemoryID& id)
        {
            return const_cast<EntitySnapshotT&>(const_cast<const EntityBase*>(this)->getSnapshot(id));
        }

        const EntitySnapshotT& getSnapshot(const MemoryID& id) const
        {
            checkEntityName(id.entityName);
            return getSnapshot(id.timestamp);
        }

        /**
         * @brief Return the snapshot with the most recent timestamp.
         * @return The latest snapshot.
         * @throw `armem::error::EntityHistoryEmpty` If the history is empty.
         * @throw `armem::error::MissingData` If the latest snapshot has no data.
         */
        EntitySnapshotT& getLatestSnapshot()
        {
            return const_cast<EntitySnapshotT&>(const_cast<const EntityBase*>(this)->getLatestSnapshot());
        }

        const EntitySnapshotT& getLatestSnapshot() const
        {
            return getLatestItem().second;
        }


        /**
         * @brief Add the given update to this entity's history.
         * @param update The update.
         * @return The snapshot ID of the update.
         */
        virtual MemoryID update(const EntityUpdate& update)
        {
            checkEntityName(update.entityID.entityName);
            this->id() = update.entityID;

            EntitySnapshotT* snapshot;

            auto it = this->_container.find(update.timeCreated);
            if (it == this->_container.end())
            {
                // Insert into history.
                snapshot = &addSnapshot(update.timeCreated);
                truncateHistoryToSize();
            }
            else
            {
                snapshot = &it->second;
            }
            // Update entry.
            snapshot->update(update);

            return snapshot->id();
        }

        /**
         * @brief Add a single snapshot with data.
         * @param snapshot The snapshot.
         * @return The stored snapshot.
         */
        EntitySnapshotT& addSnapshot(const EntitySnapshotT& snapshot)
        {
            return addSnapshot(EntitySnapshotT(snapshot));
        }

        EntitySnapshotT& addSnapshot(EntitySnapshotT&& snapshot)
        {
            auto it = this->_container.emplace(snapshot.time(), std::move(snapshot)).first;
            it->second.id().setEntityID(this->id());
            return it->second;
        }

        EntitySnapshotT& addSnapshot(const Time& timestamp)
        {
            return addSnapshot(EntitySnapshotT(timestamp));
        }


        /**
         * @brief Sets the maximum history size.
         *
         * The current history is truncated if necessary.
         */
        void setMaxHistorySize(long maxSize) override
        {
            MaxHistorySize::setMaxHistorySize(maxSize);
            truncateHistoryToSize();
        }

        std::string getKeyString() const override
        {
            return this->id().entityName;
        }
        std::string getLevelName() const override
        {
            return "entity";
        }


    private:

        /// If maximum size is set, ensure `history`'s is not higher.
        void truncateHistoryToSize()
        {
            if (_maxHistorySize >= 0)
            {
                while (this->_container.size() > size_t(_maxHistorySize))
                {
                    this->_container.erase(this->_container.begin());
                }
                ARMARX_CHECK_LESS_EQUAL(this->_container.size(), _maxHistorySize);
            }
        }

        void checkEntityName(const std::string& name) const
        {
            if (name != this->name())
            {
                throw armem::error::ContainerNameMismatch(name, getLevelName(), getKeyString());
            }
        }

        /**
         * @brief Return the snapshot with the most recent timestamp.
         * @return The latest snapshot.
         * @throw `armem::error::EntityHistoryEmpty` If the history is empty.
         */
        const typename std::map<Time, EntitySnapshotT>::value_type& getLatestItem() const
        {
            if (this->_container.empty())
            {
                throw armem::error::EntityHistoryEmpty(name(), "when getting the latest snapshot.");
            }
            return *this->_container.rbegin();
        }

    };

}
