#include "EntityInstance.h"

#include <iostream>
#include <fstream>

#include "../../core/error.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <RobotAPI/libraries/aron/core/io/dataIO/converter/Converter.h>
#include <RobotAPI/libraries/aron/core/io/dataIO/visitor/Visitor.h>
#include <RobotAPI/libraries/aron/core/io/dataIO/reader/nlohmannJSON/NlohmannJSONReader.h>
#include <RobotAPI/libraries/aron/core/io/dataIO/writer/nlohmannJSON/NlohmannJSONWriter.h>

namespace armarx::armem::d_ltm
{

    EntityInstance::EntityInstance(const EntityInstance& other) :
        Base(other),
        path(other.path)
    {
    }

    bool EntityInstance::equalsDeep(const EntityInstance& other) const
    {
        return id() == other.id();
    }



    void EntityInstance::update(const EntityUpdate& update, int index)
    {
        ARMARX_CHECK_FITS_SIZE(index, update.instancesData.size());

        this->index() = index;
    }

    EntityInstance EntityInstance::copy() const
    {
        EntityInstance d;
        this->_copySelf(d);
        return d;
    }

    void EntityInstance::_copySelf(EntityInstance& other) const
    {
        EntityInstanceBase<EntityInstance>::_copySelf(other);
    }

    std::filesystem::path EntityInstance::_fullPath() const
    {
        if (path)
        {
            return _fullPath(*path);
        }
        return std::filesystem::path();
    }

    std::filesystem::path EntityInstance::_fullPath(const std::filesystem::path& path) const
    {
        return path / id().memoryName / id().coreSegmentName / id().providerSegmentName / id().entityName / std::to_string(id().timestamp.toMicroSeconds()) / std::to_string(id().instanceIndex);
    }

    wm::EntityInstance EntityInstance::convert(const aron::typenavigator::NavigatorPtr& expectedStructure) const
    {
        std::filesystem::path p = _fullPath();
        std::filesystem::path d = p / (std::string(DATA_FILENAME) + ".json");

        if (std::filesystem::is_regular_file(d))
        {
            std::ifstream ifs(d);
            std::string file_content((std::istreambuf_iterator<char>(ifs)), (std::istreambuf_iterator<char>()));
            aron::dataIO::reader::NlohmannJSONReader dataReader(file_content);
            aron::dataIO::writer::NavigatorWriter navWriter;

            aron::dataIO::Converter::ReadAndConvert(dataReader, navWriter, expectedStructure);

            aron::datanavigator::DictNavigatorPtr aron = aron::datanavigator::DictNavigator::DynamicCastAndCheck(navWriter.getResult());
            return unwrapData(aron);
        }
        else
        {
            throw error::ArMemError("An diskMemory EntityInstance is not leading to a regular file.");
        }
    }

    void EntityInstance::reload(const std::shared_ptr<std::filesystem::path>& p_ptr)
    {
        if (!p_ptr)
        {
            ARMARX_WARNING << "The entered is NULL.";
        }
        std::filesystem::path p = _fullPath(*p_ptr);
        if (!std::filesystem::is_directory(p))
        {
            ARMARX_ERROR << "The entered path is not leading to a file! This is an error since if the folder for an EntityInstance exists there must be a data file in it (containing at least the metadata).";
        }
        else
        {
            path = p_ptr;
        }
    }

    void EntityInstance::setTo(const wm::EntityInstance& m)
    {
        std::filesystem::path p = _fullPath();
        std::filesystem::create_directories(p);

        std::filesystem::path d = p / (std::string(DATA_FILENAME) + ".json");

        if (std::filesystem::is_regular_file(d))
        {
            std::filesystem::remove(d);
        }

        std::ofstream ofs;
        ofs.open(d);

        aron::datanavigator::DictNavigatorPtr aron = wrapData(m);
        aron::dataIO::writer::NlohmannJSONWriter dataWriter;
        aron::dataIO::Visitor::VisitAndSetup(dataWriter, aron);
        std::string new_file_full_content = dataWriter.getResult().dump(2);

        ofs << new_file_full_content;
        ofs.close();
    }


    wm::EntityInstance EntityInstance::unwrapData(const aron::datanavigator::DictNavigatorPtr& dataWrapped) const
    {
        wm::EntityInstance e(id());
        wm::EntityInstanceMetadata& metadata = e.metadata();

        if (dataWrapped->hasElement(DATA_WRAPPER_DATA_FIELD))
        {
            aron::datanavigator::DictNavigatorPtr data = aron::datanavigator::DictNavigator::DynamicCastAndCheck(dataWrapped->getElement(DATA_WRAPPER_DATA_FIELD));
            e.setData(data);
        }

        auto timeCreated = aron::datanavigator::LongNavigator::DynamicCastAndCheck(dataWrapped->getElement(DATA_WRAPPER_TIME_CREATED_FIELD));
        metadata.timeCreated = Time::microSeconds(timeCreated->toAronLongPtr()->value);

        auto timeSent = aron::datanavigator::LongNavigator::DynamicCastAndCheck(dataWrapped->getElement(DATA_WRAPPER_TIME_SENT_FIELD));
        metadata.timeSent = Time::microSeconds(timeSent->toAronLongPtr()->value);

        auto timeArrived = aron::datanavigator::LongNavigator::DynamicCastAndCheck(dataWrapped->getElement(DATA_WRAPPER_TIME_ARRIVED_FIELD));
        metadata.timeArrived = Time::microSeconds(timeArrived->toAronLongPtr()->value);

        auto confidence = aron::datanavigator::DoubleNavigator::DynamicCastAndCheck(dataWrapped->getElement(DATA_WRAPPER_CONFIDENCE_FIELD));
        metadata.confidence = static_cast<float>(confidence->toAronDoublePtr()->value);

        return e;
    }

    aron::datanavigator::DictNavigatorPtr EntityInstance::wrapData(const wm::EntityInstance& e) const
    {
        auto dataWrapped = std::make_shared<aron::datanavigator::DictNavigator>();
        if (e.data())
        {
            dataWrapped->addElement(DATA_WRAPPER_DATA_FIELD, e.data());
        }

        auto timeWrapped = std::make_shared<aron::datanavigator::LongNavigator>();
        timeWrapped->setValue(Time::now().toMicroSeconds());
        dataWrapped->addElement(DATA_WRAPPER_TIME_STORED_FIELD, timeWrapped);

        const wm::EntityInstanceMetadata& metadata = e.metadata();
        auto timeCreated = std::make_shared<aron::datanavigator::LongNavigator>();
        timeCreated->setValue(metadata.timeCreated.toMicroSeconds());
        dataWrapped->addElement(DATA_WRAPPER_TIME_CREATED_FIELD, timeCreated);

        auto timeSent = std::make_shared<aron::datanavigator::LongNavigator>();
        timeSent->setValue(metadata.timeSent.toMicroSeconds());
        dataWrapped->addElement(DATA_WRAPPER_TIME_SENT_FIELD, timeSent);

        auto timeArrived = std::make_shared<aron::datanavigator::LongNavigator>();
        timeArrived->setValue(metadata.timeArrived.toMicroSeconds());
        dataWrapped->addElement(DATA_WRAPPER_TIME_ARRIVED_FIELD, timeArrived);

        auto confidence = std::make_shared<aron::datanavigator::DoubleNavigator>();
        confidence->setValue(static_cast<double>(metadata.confidence));
        dataWrapped->addElement(DATA_WRAPPER_CONFIDENCE_FIELD, confidence);

        return dataWrapped;
    }
}
