/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include "ArmarXCore/core/logging/Logging.h"
#include <vector>
#include <optional>

#include <RobotAPI/libraries/armem/core/workingmemory/Entity.h>
#include <RobotAPI/libraries/armem/core/workingmemory/EntityInstance.h>
#include <RobotAPI/libraries/aron/core/codegenerator/codeWriter/cpp/AronCppClass.h>

namespace armarx::armem
{

    /**
     * @brief Tries to cast a armem::EntityInstance to AronClass
     *
     * @tparam AronClass class name. Needs to be derived from armarx::aron::cppcodegenerator::AronCppClass
     * @param item
     * @return std::optional<AronClass>
     */
    template <typename AronClass>
    std::optional<AronClass> tryCast(const wm::EntityInstance& item)
    {
        static_assert(std::is_base_of<armarx::aron::cppcodegenerator::AronCppClass,
                      AronClass>::value);

        try
        {
            AronClass t;
            t.fromAron(item.data());
            return t;
        }
        catch (const armarx::aron::error::AronException&)
        {
            return std::nullopt;
        }
    }

    /**
     * @brief Returns all entities that can be cast to AronClass
     *
     * @tparam AronClass class name. Needs to be derived from armarx::aron::cppcodegenerator::AronCppClass
     * @param entities collection of entities
     * @return std::vector<AronClass>
     */
    template <typename AronClass>
    std::vector<AronClass>
    allOfType(const std::map<std::string, wm::Entity>& entities)
    {
        static_assert(std::is_base_of<armarx::aron::cppcodegenerator::AronCppClass,
                      AronClass>::value);

        std::vector<AronClass> outV;

        // loop over all entities and their snapshots
        for (const auto &[s, entity] : entities)
        {
            for (const auto &[ss, entitySnapshot] : entity.history())
            {
                for (const auto& entityInstance : entitySnapshot.instances())
                {
                    const auto o = tryCast<AronClass>(entityInstance);

                    if (o)
                    {
                        outV.push_back(*o);
                    }
                }
            }
        }

        return outV;
    }

    /**
     * @brief filter + transform for entities.
     *
     * Can be used instead of
     *
     *      std::vector<Bar> ret;
     *
     *      const auto allOf = allOfType<Foo>(entities);
     *      std::transform(allOf.begin(), allOf.end(), std::back_inserter(ret), pred)
     *
     * This function has the benefit that the transform function will be applied directly.
     * No intermediate vector has to be created (e.g. "allOf" in the example above).
     *
     * @tparam AronClass class name. Needs to be derived from armarx::aron::cppcodegenerator::AronCppClass
     * @param entities collection of entities
     * @param pred binary predicate function, applied to all entity instances
     * @return vector of "pred"-transformed elements that can be cast to AronClass
     */
    template <typename AronClass>
    auto transformAllOfType(const std::map<std::string, wm::Entity>& entities,
                            auto pred) -> std::vector<decltype(pred(AronClass()))>
    {
        static_assert(std::is_base_of<armarx::aron::cppcodegenerator::AronCppClass,
                      AronClass>::value);

        std::vector<decltype(pred(AronClass()))> outV;

        if (entities.empty())
        {
            ARMARX_WARNING << "No entities!";
        }

        // loop over all entities and their snapshots
        for (const auto &[s, entity] : entities)
        {
            if (entity.history().empty())
            {
                ARMARX_WARNING << "Empty history for " << s;
            }

            ARMARX_DEBUG << "History size: " << entity.history().size();

            for (const auto &[ss, entitySnapshot] : entity.history())
            {
                for (const auto& entityInstance : entitySnapshot.instances())
                {
                    const auto o = tryCast<AronClass>(entityInstance);

                    if (o)
                    {
                        outV.push_back(pred(*o));
                    }
                }
            }
        }

        return outV;
    }

} // namespace armarx::armem
