#include "Reader.h"

#include <mutex>
#include <optional>

#include "ArmarXCore/core/logging/Logging.h"
#include <ArmarXCore/core/PackagePath.h>

#include "RobotAPI/libraries/armem/core/Time.h"
#include "RobotAPI/libraries/armem/client/query/Builder.h"
#include "RobotAPI/libraries/armem_robot/robot_conversions.h"
#include "RobotAPI/libraries/armem_robot/aron_conversions.h"
#include <RobotAPI/libraries/armem_robot/aron/Robot.aron.generated.h>


namespace fs = ::std::filesystem;

namespace armarx::armem::articulated_object
{

    Reader::Reader(armem::ClientReaderComponentPluginUser& component) : component(component) {}

    std::optional<ArticulatedObject> Reader::get(const std::string& name, const armem::Time& timestamp)
    {
        const auto description = queryDescription(name, timestamp);

        if (not description)
        {
            ARMARX_ERROR << "Unknown object " << name;
            return std::nullopt;
        }

        return get(*description, timestamp);
    }


    ArticulatedObject Reader::get(const ArticulatedObjectDescription& description,
                                  const armem::Time& timestamp)
    {
        ArticulatedObject obj
        {
            .description = description,
            .instance = "", // TODO(fabian.reister):
            .config = {}, // will be populated by synchronize
            .timestamp = timestamp
        };

        synchronize(obj, timestamp);

        return obj;
    }

    void Reader::synchronize(ArticulatedObject& obj, const armem::Time& timestamp)
    {
        auto state = queryState(obj.description, timestamp);

        if (not state) /* c++20 [[unlikely]] */
        {
            ARMARX_WARNING << "Could not synchronize object " << obj.description.name;
            return;
        }

        obj.config = std::move(*state);
    }

    std::optional<robot::RobotDescription> Reader::queryDescription(const std::string& name, const armem::Time& timestamp)
    {
        // Query all entities from provider.
        armem::client::query::Builder qb;

        // clang-format off
        qb
        .coreSegments().withName(properties.coreClassSegmentName)
        .providerSegments().all() // TODO(fabian.reister): think about this: which authority is trustworthy?
        .entities().withName(name)
        .snapshots().atTime(timestamp);
        // clang-format on

        const armem::client::QueryResult qResult = memoryReader.query(qb.buildQueryInput());

        ARMARX_DEBUG << "Lookup result in reader: " << qResult;

        if (not qResult.success) /* c++20 [[unlikely]] */
        {
            return std::nullopt;
        }

        return getRobotDescription(qResult.memory);
    }

    std::optional<robot::RobotState> Reader::queryState(const robot::RobotDescription& description, const armem::Time& timestamp)
    {
        // TODO(fabian.reister): how to deal with multiple providers?

        // Query all entities from provider.
        armem::client::query::Builder qb;

        // clang-format off
        qb
        .coreSegments().withName(properties.coreInstanceSegmentName)
        .providerSegments().withName(properties.providerName) // agent
        .entities().withName(description.name)
        .snapshots().atTime(timestamp);
        // clang-format on

        const armem::client::QueryResult qResult = memoryReader.query(qb.buildQueryInput());

        ARMARX_DEBUG << "Lookup result in reader: " << qResult;

        if (not qResult.success) /* c++20 [[unlikely]] */
        {
            return std::nullopt;
        }

        return getRobotState(qResult.memory);
    }


    std::optional<robot::RobotState> Reader::getRobotState(const armarx::armem::wm::Memory& memory) const
    {
        // clang-format off
        const armem::wm::ProviderSegment& providerSegment = memory
                .getCoreSegment(properties.coreInstanceSegmentName)
                .getProviderSegment(properties.providerName);
        // clang-format on
        const auto entities = simox::alg::get_values(providerSegment.entities());

        if (entities.empty())
        {
            ARMARX_WARNING << "No entity found";
            return std::nullopt;
        }

        const auto entitySnapshots = simox::alg::get_values(entities.front().history());

        if (entitySnapshots.empty())
        {
            ARMARX_WARNING << "No entity snapshots found";
            return std::nullopt;
        }

        // TODO(fabian.reister): check if 0 available
        const armem::wm::EntityInstance& instance = entitySnapshots.front().getInstance(0);

        return robot::convertRobotState(instance);
    }



    std::optional<robot::RobotDescription> Reader::getRobotDescription(const armarx::armem::wm::Memory& memory) const
    {
        // clang-format off
        const armem::wm::ProviderSegment& providerSegment = memory
                .getCoreSegment(properties.coreClassSegmentName)
                .getProviderSegment(properties.providerName); // TODO(fabian.reister): all
        // clang-format on
        const auto entities = simox::alg::get_values(providerSegment.entities());

        if (entities.empty())
        {
            ARMARX_WARNING << "No entity found";
            return std::nullopt;
        }

        const auto entitySnapshots = simox::alg::get_values(entities.front().history());

        if (entitySnapshots.empty())
        {
            ARMARX_WARNING << "No entity snapshots found";
            return std::nullopt;
        }

        // TODO(fabian.reister): check if 0 available
        const armem::wm::EntityInstance& instance = entitySnapshots.front().getInstance(0);

        return robot::convertRobotDescription(instance);
    }

} // namespace armarx::armem::articulated_object