#include "Segment.h"

#include <sstream>

#include <SimoxUtility/algorithm/get_map_keys_values.h>

#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/libraries/core/remoterobot/RemoteRobot.h>

#include <RobotAPI/libraries/aron/common/aron_conversions.h>

#include <RobotAPI/libraries/armem/core/aron_conversions.h>
#include <RobotAPI/libraries/armem/client/Writer.h>
#include <RobotAPI/libraries/armem/client/query/Builder.h>
#include <RobotAPI/libraries/armem/client/query/query_fns.h>
#include <RobotAPI/libraries/armem/aron/MemoryID.aron.generated.h>

#include <RobotAPI/libraries/armem_robot/aron/Robot.aron.generated.h>
#include <RobotAPI/libraries/armem_robot/aron/RobotDescription.aron.generated.h>
#include <RobotAPI/libraries/armem_robot/aron_conversions.h>
#include <RobotAPI/libraries/armem_robot/robot_conversions.h>

#include <RobotAPI/libraries/armem_objects/server/articulated_object_class/Segment.h>

#include "Visu.h"


namespace armarx::armem::server::obj::articulated_object_instance
{

    Segment::Segment(armem::server::MemoryToIceAdapter& memoryToIceAdapter, std::mutex& memoryMutex) :
        iceMemory(memoryToIceAdapter),
        memoryMutex(memoryMutex)
    {
        Logging::setTag("ArticulatedObjectInstanceSegment");
    }

    Segment::~Segment() = default;

    void Segment::defineProperties(armarx::PropertyDefinitionsPtr& defs, const std::string& prefix)
    {
        defs->optional(p.coreInstanceSegmentName, prefix + "CoreSegmentName", "Name of the object instance core segment.");
        defs->optional(p.maxHistorySize, prefix + "MaxHistorySize", "Maximal size of object poses history (-1 for infinite).");
    }

    void Segment::init()
    {
        ARMARX_CHECK_NOT_NULL(iceMemory.workingMemory);

        coreSegment = &iceMemory.workingMemory->addCoreSegment(p.coreInstanceSegmentName, arondto::Robot::toInitialAronType());
        coreSegment->setMaxHistorySize(p.maxHistorySize);

    }

    void Segment::connect(viz::Client arviz)
    {
        this->visu = std::make_unique<Visu>(arviz, *this);
        visu->init();
    }

    void Segment::setArticulatedObjectClassSegment(const articulated_object_class::Segment& segment)
    {
        classSegment = &segment;
    }

    ::armarx::armem::articulated_object::ArticulatedObjects Segment::getArticulatedObjects() const
    {
        ARMARX_CHECK_NOT_NULL(classSegment);
        const auto knownObjectClasses = classSegment->getKnownObjectClasses();

        const auto resolveDescriptionLink = [&](auto & articulatedObject, const auto & aronDescriptionLink) -> bool
        {
            armem::MemoryID descriptionLink;
            fromAron(aronDescriptionLink, descriptionLink);

            ARMARX_DEBUG << "Lookup key is " << descriptionLink;

            // const auto keys = simox::alg::get_keys(knownObjectClasses);
            // ARMARX_DEBUG << "Known keys " << keys;

            const auto it = knownObjectClasses.find(descriptionLink);
            if (it == knownObjectClasses.end())
            {
                ARMARX_WARNING << "Unknown object class " ; //<< aronArticulatedObject.description;
                return false;
            }

            articulatedObject.description = it->second;
            return true;
        };

        ::armarx::armem::articulated_object::ArticulatedObjects objects;
        for (const auto& [_, provSeg] : iceMemory.workingMemory->getCoreSegment(p.coreInstanceSegmentName))
        {
            for (const auto& [_, entity] :  provSeg.entities())
            {
                const auto& entityInstance = entity.getLatestSnapshot().getInstance(0);

                arondto::Robot aronArticulatedObject;
                aronArticulatedObject.fromAron(entityInstance.data());

                armem::articulated_object::ArticulatedObject articulatedObject;
                fromAron(aronArticulatedObject, articulatedObject);

                // resolve memory link for description
                const arondto::MemoryID& aronDescriptionLink = aronArticulatedObject.description;
                if (not resolveDescriptionLink(articulatedObject, aronDescriptionLink))
                {
                    continue;
                }

                objects.push_back(articulatedObject);
            }
        }

        return objects;
    }



    // void Segment::RemoteGui::setup(const Segment& data)
    // {
    //     using namespace armarx::RemoteGui::Client;

    //     maxHistorySize.setValue(std::max(1, int(data.p.maxHistorySize)));
    //     maxHistorySize.setRange(1, 1e6);
    //     infiniteHistory.setValue(data.p.maxHistorySize == -1);
    //     discardSnapshotsWhileAttached.setValue(data.p.discardSnapshotsWhileAttached);

    //     GridLayout grid;
    //     int row = 0;
    //     grid.add(Label("Max History Size"), {row, 0}).add(maxHistorySize, {row, 1});
    //     row++;
    //     grid.add(Label("Infinite History Size"), {row, 0}).add(infiniteHistory, {row, 1});
    //     row++;
    //     grid.add(Label("Discard Snapshots while Attached"), {row, 0}).add(discardSnapshotsWhileAttached, {row, 1});
    //     row++;

    //     group.setLabel("Data");
    //     group.addChild(grid);
    // }

    // void Segment::RemoteGui::update(Segment& data)
    // {
    //     if (infiniteHistory.hasValueChanged() || maxHistorySize.hasValueChanged()
    //         || discardSnapshotsWhileAttached.hasValueChanged())
    //     {
    //         std::scoped_lock lock(data.memoryMutex);

    //         if (infiniteHistory.hasValueChanged() || maxHistorySize.hasValueChanged())
    //         {
    //             data.p.maxHistorySize = infiniteHistory.getValue() ? -1 : maxHistorySize.getValue();
    //             if (data.coreSegment)
    //             {
    //                 data.coreSegment->setMaxHistorySize(long(data.p.maxHistorySize));
    //             }
    //         }

    //         data.p.discardSnapshotsWhileAttached = discardSnapshotsWhileAttached.getValue();
    //     }
    // }

}  // namespace armarx::armem::server::obj::articulated_object_instance
