/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "Pair.h"


namespace armarx::aron::cppcodegenerator::serializer
{
    PairSerializer::PairSerializer(const typenavigator::PairNavigatorPtr& e) :
        Serializer("std::pair<" + ExtractCppTypename(e->getFirstAcceptedType()) + ", " + ExtractCppTypename(e->getSecondAcceptedType()) + ">", simox::meta::get_type_name(typeid(data::AronList)), simox::meta::get_type_name(typeid(type::AronPair))),
        typenavigator(e)
    {

    }

    std::vector<CppFieldPtr> PairSerializer::getPublicVariableDeclarations(const std::string& name) const
    {
        CppFieldPtr field = CppFieldPtr(new CppField(getFullCppTypename(), name));
        return {field};
    }

    std::vector<std::pair<std::string, std::string>> PairSerializer::getCtorInitializers(const std::string&) const
    {
        return {};
    }

    CppBlockPtr PairSerializer::getCtorBlock(const std::string&) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        return b;
    }

    CppBlockPtr PairSerializer::getInitializeBlock(const std::string& accessor) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());

        auto child_s1 = FromAronTypeNaviagtorPtr(typenavigator->getFirstAcceptedType());
        CppBlockPtr b21 = child_s1->getInitializeBlock(accessor + ".first");
        b->appendBlock(b21);

        auto child_s2 = FromAronTypeNaviagtorPtr(typenavigator->getSecondAcceptedType());
        CppBlockPtr b22 = child_s2->getInitializeBlock(accessor + ".second");
        b->appendBlock(b22);

        return b;
    }

    CppBlockPtr PairSerializer::getWriteInitialTypeBlock(const std::string& accessor) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine("w.writeStartPair();");

        std::string escaped_accessor = EscapeAccessor(accessor);

        auto child_s1 = FromAronTypeNaviagtorPtr(typenavigator->getFirstAcceptedType());
        std::string accessor_iterator1 = escaped_accessor + PAIR_FIRST_ACCESSOR;
        b->addLine("w.writeKey(\"0\");");
        CppBlockPtr b21 = child_s1->getWriteInitialTypeBlock(accessor_iterator1);
        b->appendBlock(b21);

        auto child_s2 = FromAronTypeNaviagtorPtr(typenavigator->getSecondAcceptedType());
        std::string accessor_iterator2 = escaped_accessor + PAIR_SECOND_ACCESSOR;
        b->addLine("w.writeKey(\"1\");");
        CppBlockPtr b22 = child_s2->getWriteInitialTypeBlock(accessor_iterator2);
        b->appendBlock(b22);

        b->addLine("w.writeEndPair();");
        return b;
    }

    CppBlockPtr PairSerializer::getWriteCurrentTypeBlock(const std::string& accessor) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine("w.writeStartPair();");

        std::string escaped_accessor = EscapeAccessor(accessor);

        auto child_s1 = FromAronTypeNaviagtorPtr(typenavigator->getFirstAcceptedType());
        std::string accessor_iterator1 = escaped_accessor + PAIR_FIRST_ACCESSOR;
        b->addLine("w.writeKey(\"0\");");
        CppBlockPtr b21 = child_s1->getWriteCurrentTypeBlock(accessor_iterator1);
        b->appendBlock(b21);

        auto child_s2 = FromAronTypeNaviagtorPtr(typenavigator->getSecondAcceptedType());
        std::string accessor_iterator2 = escaped_accessor + PAIR_SECOND_ACCESSOR;
        b->addLine("w.writeKey(\"1\");");
        CppBlockPtr b22 = child_s2->getWriteCurrentTypeBlock(accessor_iterator2);
        b->appendBlock(b22);

        b->addLine("w.writeEndPair();");
        return b;
    }

    CppBlockPtr PairSerializer::getResetBlock(const std::string& accessor) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());

        auto child_s1 = FromAronTypeNaviagtorPtr(typenavigator->getFirstAcceptedType());
        CppBlockPtr b21 = child_s1->getResetBlock(accessor + ".first");
        b->appendBlock(b21);

        auto child_s2 = FromAronTypeNaviagtorPtr(typenavigator->getSecondAcceptedType());
        CppBlockPtr b22 = child_s2->getResetBlock(accessor + ".second");
        b->appendBlock(b22);

        return b;
    }

    CppBlockPtr PairSerializer::getWriteBlock(const std::string& accessor) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine("w.writeStartList();");

        auto child_s1 = FromAronTypeNaviagtorPtr(typenavigator->getFirstAcceptedType());
        CppBlockPtr b21 = child_s1->getWriteBlock(accessor + ".first");
        b->appendBlock(b21);

        auto child_s2 = FromAronTypeNaviagtorPtr(typenavigator->getSecondAcceptedType());
        CppBlockPtr b22 = child_s2->getWriteBlock(accessor + ".second");
        b->appendBlock(b22);

        b->addLine("w.writeEndList();");
        return b;
    }

    CppBlockPtr PairSerializer::getReadBlock(const std::string& accessor) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine(accessor + " = {};");
        b->addLine("r.readStartList();");

        auto child_s1 = FromAronTypeNaviagtorPtr(typenavigator->getFirstAcceptedType());
        CppBlockPtr b21 = child_s1->getReadBlock(accessor + ".first");
        b->appendBlock(b21);

        auto child_s2 = FromAronTypeNaviagtorPtr(typenavigator->getSecondAcceptedType());
        CppBlockPtr b22 = child_s2->getReadBlock(accessor + ".second");
        b->appendBlock(b22);

        b->addLine("r.readEndList();");
        return b;
    }

    CppBlockPtr PairSerializer::getEqualsBlock(const std::string& accessor, const std::string& otherInstanceAccessor) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine("if (not (" + accessor + " == " + otherInstanceAccessor + "))");
        b->addLine("\t return false;");
        return b;
    }
}

